from django.utils.functional import cached_property

from .settings import api_token_auth_settings


class UserAuthorization(object):
    def __init__(self, user, api_token_payload, settings=None):
        """
        Initialize authorization info from user and API Token payload.
        """
        self.user = user
        self.data = api_token_payload
        self.settings = settings or api_token_auth_settings

    def has_api_scopes(self, *api_scopes):
        """
        Test if all given API scopes are authorized.

        :type api_scopes: list[str]
        :param api_scopes: The API scopes to test

        :rtype: bool|None
        :return:
          True or False, if the API Token has the API scopes field set,
          otherwise None
        """
        if self._authorized_api_scopes is None:
            return None
        return all((x in self._authorized_api_scopes) for x in api_scopes)

    def has_api_scope_with_prefix(self, prefix):
        """
        Test if there is an API scope with the given prefix.

        :rtype: bool|None
        """
        if self._authorized_api_scopes is None:
            return None
        return any(
            x == prefix or x.startswith(prefix + '.')
            for x in self._authorized_api_scopes)

    @cached_property
    def _authorized_api_scopes(self):
        api_scopes = self.data.get(self.settings.API_AUTHORIZATION_FIELD)
        return (set(api_scopes)
                if is_list_of_non_empty_strings(api_scopes) else None)


def is_list_of_non_empty_strings(value):
    if not isinstance(value, list):
        return False
    return all(isinstance(x, str) and x for x in value)



from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext as _
from django.db import transaction, IntegrityError
from rest_framework import exceptions


def oidc_to_user_data(payload):
    """
    Map OIDC claims to Django user fields.
    """
    payload = payload.copy()

    field_map = {
        'given_name': 'first_name',
        'family_name': 'last_name',
        'email': 'email',
    }
    ret = {}
    for token_attr, user_attr in field_map.items():
        if token_attr not in payload:
            continue
        ret[user_attr] = payload.pop(token_attr)
    ret.update(payload)

    return ret


def populate_user(user, data):
    exclude_fields = ['is_staff', 'password', 'is_superuser', 'id']
    user_fields = [f.name for f in user._meta.fields
                   if f.name not in exclude_fields]
    changed = False
    for field in user_fields:
        if field in data:
            val = data[field]
            if getattr(user, field) != val:
                setattr(user, field, val)
                changed = True

    return changed


def update_user(user, payload, oidc=False):
    if oidc:
        payload = oidc_to_user_data(payload)

    changed = populate_user(user, payload)
    if changed or not user.pk:
        user.save()

    ad_groups = payload.get('ad_groups', None)
    # Only update AD groups if it's a list of non-empty strings
    if isinstance(ad_groups, list) and (
            all([isinstance(x, str) and x for x in ad_groups])):
        user.update_ad_groups(ad_groups)


def _try_create_or_update(user_id, payload, oidc):
    user_model = get_user_model()
    with transaction.atomic():
        try:
            user = user_model.objects.select_for_update().get(uuid=user_id)
        except user_model.DoesNotExist:
            user = user_model(uuid=user_id)
            user.set_unusable_password()
        update_user(user, payload, oidc)
    return user


def get_or_create_user(payload, oidc=False):
    user_id = payload.get('sub')
    if not user_id:
        msg = _('Invalid payload.')
        raise exceptions.AuthenticationFailed(msg)

    try_again = False
    try:
        user = _try_create_or_update(user_id, payload, oidc)
    except IntegrityError:
        # If we get an integrity error, it probably meant a race
        # condition with another process. Another attempt should
        # succeed.
        try_again = True
    if try_again:
        # We try again without catching exceptions this time.
        user = _try_create_or_update(user_id, payload, oidc)

    # If allauth.socialaccount is installed, create the SocialAcount
    # that corresponds to this user. Otherwise logins through
    # allauth will not work for the user later on.
    if 'allauth.socialaccount' in settings.INSTALLED_APPS:
        from allauth.socialaccount.models import SocialAccount, EmailAddress

        if oidc:
            provider_name = 'helsinki_oidc'
        else:
            provider_name = 'helsinki'
        args = {'provider': provider_name, 'uid': user_id}
        try:
            account = SocialAccount.objects.get(**args)
            assert account.user_id == user.id
        except SocialAccount.DoesNotExist:
            account = SocialAccount(**args)
            account.extra_data = payload
            account.user = user
            account.save()

            try:
                email = EmailAddress.objects.get(email__iexact=user.email)
                assert email.user == user
            except EmailAddress.DoesNotExist:
                email = EmailAddress(email=user.email.lower(), primary=True,
                                     user=user, verified=True)
                email.save()

    return user



from allauth.socialaccount import providers
from allauth.socialaccount.providers.base import ProviderAccount
from allauth.socialaccount.providers.oauth2.provider import OAuth2Provider

from helusers.user_utils import oidc_to_user_data
from helusers.utils import uuid_to_username


class HelsinkiOIDCAccount(ProviderAccount):
    def get_profile_url(self):
        return self.account.extra_data.get('html_url')

    def get_avatar_url(self):
        return self.account.extra_data.get('avatar_url')

    def to_str(self):
        dflt = super(HelsinkiOIDCAccount, self).to_str()
        return self.account.extra_data.get('name', dflt)


class HelsinkiOIDCProvider(OAuth2Provider):
    id = 'helsinki_oidc'
    name = 'City of Helsinki employees (OIDC)'
    package = 'helusers.providers.helsinki_oidc'
    account_class = HelsinkiOIDCAccount

    def extract_uid(self, data):
        return str(data['sub'])

    def extract_common_fields(self, data):
        ret = oidc_to_user_data(data)
        ret['username'] = uuid_to_username(data['sub'])
        return ret

    def get_default_scope(self):
        return ['openid profile email']


providers.registry.register(HelsinkiOIDCProvider)



import requests
from django.utils.encoding import smart_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _
from oidc_auth.authentication import JSONWebTokenAuthentication
from oidc_auth.util import cache
from rest_framework.authentication import get_authorization_header
from rest_framework.exceptions import AuthenticationFailed

from .authz import UserAuthorization
from .settings import api_token_auth_settings
from .user_utils import get_or_create_user


class ApiTokenAuthentication(JSONWebTokenAuthentication):
    def __init__(self, settings=None, **kwargs):
        self.settings = settings or api_token_auth_settings
        super(ApiTokenAuthentication, self).__init__(**kwargs)

    @cached_property
    def auth_scheme(self):
        return self.settings.AUTH_SCHEME or 'Bearer'

    @property
    def oidc_config(self):
        return self.get_oidc_config()

    @cache(ttl=api_token_auth_settings.OIDC_CONFIG_EXPIRATION_TIME)
    def get_oidc_config(self):
        url = self.settings.ISSUER + '/.well-known/openid-configuration'
        return requests.get(url).json()

    def authenticate(self, request):
        jwt_value = self.get_jwt_value(request)
        if jwt_value is None:
            return None

        payload = self.decode_jwt(jwt_value)
        self.validate_claims(payload)

        user_resolver = self.settings.USER_RESOLVER  # Default: resolve_user
        user = user_resolver(request, payload)
        auth = UserAuthorization(user, payload, self.settings)

        if self.settings.REQUIRE_API_SCOPE_FOR_AUTHENTICATION:
            api_scope = self.settings.API_SCOPE_PREFIX
            if not auth.has_api_scope_with_prefix(api_scope):
                raise AuthenticationFailed(
                    _("Not authorized for API scope \"{api_scope}\"")
                    .format(api_scope=api_scope))

        return (user, auth)

    def get_jwt_value(self, request):
        auth = get_authorization_header(request).split()

        if not auth or smart_text(auth[0]).lower() != self.auth_scheme.lower():
            return None

        if len(auth) == 1:
            raise AuthenticationFailed(
                _("Invalid Authorization header. No credentials provided"))
        elif len(auth) > 2:
            raise AuthenticationFailed(
                _("Invalid Authorization header. "
                  "Credentials string should not contain spaces."))

        return auth[1]

    def authenticate_header(self, request):
        return '{auth_scheme} realm="{realm}"'.format(
            auth_scheme=self.auth_scheme,
            realm=self.www_authenticate_realm)

    def get_audiences(self, api_token):
        return {self.settings.AUDIENCE}


def resolve_user(request, payload):
    return get_or_create_user(payload, oidc=True)



from collections import OrderedDict
from urllib.parse import urlencode

from django.contrib.auth.views import LogoutView as DjangoLogoutView
from django.views.generic.base import RedirectView
from django.contrib.auth import logout, REDIRECT_FIELD_NAME
from django.http import HttpResponseRedirect
from django.contrib import messages
from django.utils.translation import gettext_lazy as _
from django.urls import reverse

LANGUAGE_FIELD_NAME = 'ui_locales'


class LogoutView(DjangoLogoutView):
    def dispatch(self, request, *args, **kwargs):
        was_authenticated = request.user.is_authenticated
        session = request.session
        end_session_url = session.get('social_auth_end_session_url')
        if end_session_url:
            self.next_page = end_session_url
        ret = super().dispatch(request, *args, **kwargs)
        if was_authenticated:
            messages.success(request, _("You have been successfully logged out."))
        return ret


class LogoutCompleteView(DjangoLogoutView):
    def dispatch(self, request, *args, **kwargs):
        return HttpResponseRedirect(self.get_next_page())


class LoginView(RedirectView):
    permanent = False

    def get_redirect_url(self, *args, **kwargs):
        # We make sure the user is logged out first, because otherwise
        # PSA will enter the connect flow where a new social authentication
        # method is connected to an existing user account.
        logout(self.request)

        url = reverse('social:begin', kwargs=dict(backend='tunnistamo'))
        redirect_to = self.request.GET.get(REDIRECT_FIELD_NAME)
        lang = self.request.GET.get(LANGUAGE_FIELD_NAME)

        query_params = OrderedDict()
        if redirect_to:
            query_params[REDIRECT_FIELD_NAME] = redirect_to
        if lang:
            query_params[LANGUAGE_FIELD_NAME] = lang
        if query_params:
            url += '?' + urlencode(query_params)

        return url



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from allauth.socialaccount.models import SocialApp
from helusers.providers.helsinki.provider import HelsinkiProvider


class Command(BaseCommand):
    help = 'Create or update helusers allauth SocialApp'

    def handle(self, *args, **options):
        changed = False
        try:
            app = SocialApp.objects.get(provider=HelsinkiProvider.id)
        except SocialApp.DoesNotExist:
            app = SocialApp(provider=HelsinkiProvider.id)
            self.stdout.write(self.style.SUCCESS('Creating new SocialApp'))

        if not app.name:
            app.name = 'Helsingin kaupungin työntekijät'
            changed = True

        client_id = secret_key = None

        jwt_settings = getattr(settings, 'JWT_AUTH')
        if jwt_settings:
            client_id = jwt_settings.get('JWT_AUDIENCE')
            secret_key = jwt_settings.get('JWT_SECRET_KEY')

        if not client_id:
            raise ImproperlyConfigured("You must set JWT_AUTH['JWT_AUDIENCE'] to correspond to your client ID")
        if not secret_key:
            raise ImproperlyConfigured("You must set JWT_AUTH['JWT_SECRET_KEY'] to correspond to your secret key")

        if app.client_id != client_id:
            changed = True
            app.client_id = client_id
        if app.secret != secret_key:
            changed = True
            app.secret = secret_key
        if changed:
            app.save()

        if not app.sites.exists():
            from django.contrib.sites.models import Site

            app.sites.add(Site.objects.get(id=settings.SITE_ID))
            changed = True

        if changed:
            self.stdout.write(self.style.SUCCESS('SocialApp successfully updated'))
        else:
            self.stdout.write(self.style.NOTICE('Already synced -- no changes needed'))



import logging
import requests
from datetime import datetime, timedelta

from django.contrib.auth import get_user_model
from django.conf import settings
from .user_utils import get_or_create_user
from .utils import uuid_to_username
from .tunnistamo_oidc import TunnistamoOIDCAuth


User = get_user_model()
logger = logging.getLogger(__name__)


def ensure_uuid_match(details, backend, response, user=None, *args, **kwargs):
    if not isinstance(backend, TunnistamoOIDCAuth):
        return

    user = user or details.get('user')
    if user is not None:
        if user.uuid != details.get('uid') or response.get('sub'):
            return {'user': None}


def get_username(details, backend, response, *args, **kwargs):
    """Sets the `username` argument.

    If the user exists already, use the existing username. Otherwise
    generate username from the `new_uuid` using the
    `helusers.utils.uuid_to_username` function.
    """

    user = details.get('user')
    if not user:
        user_uuid = kwargs.get('uid')
        if not user_uuid:
            return

        username = uuid_to_username(user_uuid)
    else:
        username = user.username

    return {
        'username': username
    }


def create_or_update_user(details, backend, response, user=None, *args, **kwargs):
    response = response.copy()
    username = kwargs.get('username')
    if username:
        response['username'] = username

    user = get_or_create_user(response, oidc=True)
    return {
        'user': user
    }


def store_end_session_url(details, backend, response, user=None, *args, **kwargs):
    if not user or not user.is_authenticated:
        return

    if not hasattr(backend, 'get_end_session_url'):
        return
    request = kwargs['request']
    if not request:
        return

    end_session_url = backend.get_end_session_url(request, response['id_token'])
    if not end_session_url:
        return

    request.session['social_auth_end_session_url'] = end_session_url
    if 'id_token' in response:
        request.session['social_auth_id_token'] = response['id_token']


def fetch_api_tokens(details, backend, response, user=None, social=None, request=None, *args, **kwargs):
    if not isinstance(backend, TunnistamoOIDCAuth):
        return
    if not user or not user.is_authenticated or not social or not request:
        return

    scopes = set(backend.get_scope())
    # slightly ghetto-style filtering, but it works for now
    api_scopes = {s for s in scopes if s.startswith('https://')}
    tunnistamo_scopes = scopes - api_scopes

    request.session['access_token_scope'] = list(tunnistamo_scopes)
    request.session['access_token'] = response['access_token']
    expires_at = datetime.now() + timedelta(seconds=response['expires_in'])
    request.session['access_token_expires_at'] = expires_at
    expires_at_ts = int(expires_at.timestamp()) * 1000
    request.session['access_token_expires_at_ts'] = expires_at_ts

    if not api_scopes:
        return

    logger.info('Retrieving API tokens for scopes %s' % ' '.join(api_scopes))

    headers = {'Authorization': 'Bearer %s' % social.extra_data['access_token']}
    url = settings.TUNNISTAMO_BASE_URL + '/api-tokens/'
    resp = requests.post(url, headers=headers)
    if resp.status_code != 200:
        logger.error('Unable to get API tokens: HTTP %d' % (resp.status_code))
        return
    request.session['api_tokens'] = resp.json()



import requests

from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter, OAuth2LoginView, OAuth2CallbackView
)
from .provider import HelsinkiOIDCProvider


class HelsinkiOIDCOAuth2Adapter(OAuth2Adapter):
    provider_id = HelsinkiOIDCProvider.id
    access_token_url = 'https://api.hel.fi/sso/openid/token/'
    authorize_url = 'https://api.hel.fi/sso/openid/authorize/'
    profile_url = 'https://api.hel.fi/sso/openid/userinfo/'

    def complete_login(self, request, app, token, **kwargs):
        headers = {'Authorization': 'Bearer {0}'.format(token.token)}
        resp = requests.get(self.profile_url, headers=headers)
        assert resp.status_code == 200
        extra_data = resp.json()
        return self.get_provider().sociallogin_from_response(request,
                                                             extra_data)


oauth2_login = OAuth2LoginView.adapter_view(HelsinkiOIDCOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(HelsinkiOIDCOAuth2Adapter)



import base64
from uuid import UUID


def uuid_to_username(uuid):
    """
    Convert UUID to username.

    >>> uuid_to_username('00fbac99-0bab-5e66-8e84-2e567ea4d1f6')
    'u-ad52zgilvnpgnduefzlh5jgr6y'

    >>> uuid_to_username(UUID('00fbac99-0bab-5e66-8e84-2e567ea4d1f6'))
    'u-ad52zgilvnpgnduefzlh5jgr6y'
    """
    uuid_data = getattr(uuid, 'bytes', None) or UUID(uuid).bytes
    b32coded = base64.b32encode(uuid_data)
    return 'u-' + b32coded.decode('ascii').replace('=', '').lower()


def username_to_uuid(username):
    """
    Convert username to UUID.

    >>> username_to_uuid('u-ad52zgilvnpgnduefzlh5jgr6y')
    UUID('00fbac99-0bab-5e66-8e84-2e567ea4d1f6')
    """
    if not username.startswith('u-') or len(username) != 28:
        raise ValueError('Not an UUID based username: %r' % (username,))
    decoded = base64.b32decode(username[2:].upper() + '======')
    return UUID(bytes=decoded)



# Import SOCIAL_AUTH_PIPELINE in your settings.py
SOCIAL_AUTH_PIPELINE = (
    # Get the information we can about the user and return it in a simple
    # format to create the user instance later. On some cases the details are
    # already part of the auth response from the provider, but sometimes this
    # could hit a provider API.
    'social_core.pipeline.social_auth.social_details',

    # Get the social uid from whichever service we're authing thru. The uid is
    # the unique identifier of the given user in the provider.
    'social_core.pipeline.social_auth.social_uid',

    # Reset logged-in user if UUID differs
    'helusers.pipeline.ensure_uuid_match',

    # Generate username from UUID
    'helusers.pipeline.get_username',

    # Checks if the current social-account is already associated in the site.
    'social_core.pipeline.social_auth.social_user',

    # Get or create the user and update user data
    'helusers.pipeline.create_or_update_user',

    # Create the record that associated the social account with this user.
    'social_core.pipeline.social_auth.associate_user',

    # Populate the extra_data field in the social record with the values
    # specified by settings (and the default ones like access_token, etc).
    'social_core.pipeline.social_auth.load_extra_data',

    # Store the end session URL in the user's session data so that
    # we can format logout links properly.
    'helusers.pipeline.store_end_session_url',

    # If the access token gives access to external APIs, fetch the
    # API tokens here.
    'helusers.pipeline.fetch_api_tokens',
)



from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from .user_utils import update_user


class SocialAccountAdapter(DefaultSocialAccountAdapter):
    def pre_social_login(self, request, sociallogin):
        """Update user based on token information."""

        user = sociallogin.user
        # If the user hasn't been saved yet, it will be updated
        # later on in the sign-up flow.
        if not user.pk:
            return

        data = sociallogin.account.extra_data
        oidc = sociallogin.account.provider == 'helsinki_oidc'
        update_user(user, data, oidc)

    def populate_user(self, request, sociallogin, data):
        user = sociallogin.user
        exclude_fields = ['is_staff', 'password', 'is_superuser', 'id']
        user_fields = [f.name for f in user._meta.fields if f not in exclude_fields]
        for field in user_fields:
            if field in data:
                setattr(user, field, data[field])
        return user

    def save_user(self, request, sociallogin, form=None):
        # This is called at the end of the new user flow.
        u = sociallogin.user
        u.set_unusable_password()
        sociallogin.save(request)

        data = sociallogin.account.extra_data
        oidc = sociallogin.account.provider == 'helsinki_oidc'
        update_user(u, data, oidc)

        return u



from django.conf import settings
from rest_framework.settings import APISettings


_user_settings = getattr(settings, 'OIDC_API_TOKEN_AUTH', None)

_defaults = dict(
    # Accepted audience, the API Token must have this in its aud field
    AUDIENCE=None,

    # API scope prefix for permission checks
    API_SCOPE_PREFIX=None,

    # Is API scope required for successful authentication.
    #
    # If this setting is set, then authentication will fail, if the API
    # scopes field doesn't contain the API_SCOPE_PREFIX or any value that
    # starts with API_SCOPE_PREFIX and a dot.
    #
    # E.g. if API_SCOPE_PREFIX='xyz' and this is set to true, then the
    # authentication will fail if the API scopes doesn't contain either
    # 'xyz' or an item that starts with 'xyz.' (like 'xyz.readonly' or
    # 'xyz.view').
    REQUIRE_API_SCOPE_FOR_AUTHENTICATION=False,

    # Field name containing the API scopes authorized by the user
    API_AUTHORIZATION_FIELD='https://api.hel.fi/auth',

    # URL of the OpenID Provider
    ISSUER='https://tunnistamo.hel.fi',

    # Auth scheme used in the Authorization header
    AUTH_SCHEME='Bearer',

    # Function for resolving users
    USER_RESOLVER='helusers.oidc.resolve_user',

    # OIDC config expiration time
    OIDC_CONFIG_EXPIRATION_TIME=24 * 60 * 60
)

_import_strings = [
    'USER_RESOLVER',
]

api_token_auth_settings = APISettings(
    _user_settings, _defaults, _import_strings)



"""URLs module"""
from django.urls import path
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from . import views


app_name = 'helusers'

urlpatterns = [
    path('logout/', views.LogoutView.as_view(), name='auth_logout'),
    path('logout/complete/', views.LogoutCompleteView.as_view(), name='auth_logout_complete'),
    path('login/', views.LoginView.as_view(), name='auth_login'),
]

if not settings.LOGOUT_REDIRECT_URL:
    raise ImproperlyConfigured("You must configure LOGOUT_REDIRECT_URL to use helusers views.")



import pytest
import random
from uuid import UUID
from helusers.utils import uuid_to_username, username_to_uuid


def test_uuid_to_username():
    assert uuid_to_username('00fbac99-0bab-5e66-8e84-2e567ea4d1f6') == 'u-ad52zgilvnpgnduefzlh5jgr6y'


def test_username_to_uuid():
    assert username_to_uuid('u-ad52zgilvnpgnduefzlh5jgr6y') == UUID('00fbac99-0bab-5e66-8e84-2e567ea4d1f6')


def test_reflective_username_uuid_relationship():
    rd = random.Random()
    rd.seed(0)

    for uuid in [UUID(int=rd.getrandbits(128)) for i in range(0,100)]:
        assert username_to_uuid(uuid_to_username(uuid)) == uuid



import django
from django.apps import apps
from django.contrib import admin
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy
from django.views.decorators.cache import never_cache


if hasattr(settings, 'SITE_TYPE'):
    if settings.SITE_TYPE not in ('dev', 'test', 'production'):
        raise ImproperlyConfigured("SITE_TYPE must be either 'dev', 'test' or 'production'")


PROVIDERS = (
    ('helusers.providers.helsinki', 'helsinki_login'),
    ('helusers.providers.helsinki_oidc', 'helsinki_oidc_login')
)


class AdminSite(admin.AdminSite):
    login_template = "admin/hel_login.html"

    def __init__(self, *args, **kwargs):
        super(AdminSite, self).__init__(*args, **kwargs)

    @property
    def site_header(self):
        if 'django.contrib.sites' in settings.INSTALLED_APPS:
            Site = apps.get_model(app_label='sites', model_name='Site')
            site = Site.objects.get_current()
            site_name = site.name
        elif hasattr(settings, 'WAGTAIL_SITE_NAME'):
            site_name = settings.WAGTAIL_SITE_NAME
        else:
            return ugettext_lazy("Django admin")
        return ugettext_lazy("%(site_name)s admin") % {'site_name': site_name}

    def each_context(self, request):
        ret = super(AdminSite, self).each_context(request)
        ret['site_type'] = getattr(settings, 'SITE_TYPE', 'dev')
        ret['redirect_path'] = request.GET.get('next', None)
        provider_installed = False
        if 'helusers.tunnistamo_oidc.TunnistamoOIDCAuth' in settings.AUTHENTICATION_BACKENDS:
            provider_installed = True
            login_url = reverse('helusers:auth_login')
            logout_url = reverse('helusers:auth_logout')
        else:
            for provider, login_view in PROVIDERS:
                if provider not in settings.INSTALLED_APPS:
                    continue
                provider_installed = True
                login_url = reverse(login_view)
                break
            logout_url = None

        ret['helsinki_provider_installed'] = provider_installed
        if provider_installed:
            ret['helsinki_login_url'] = login_url
            ret['helsinki_logout_url'] = logout_url

        ret['grappelli_installed'] = 'grappelli' in settings.INSTALLED_APPS
        if ret['grappelli_installed']:
            ret['grappelli_admin_title'] = self.site_header
            ret['base_site_template'] = 'admin/base_site_grappelli.html'
        else:
            ret['base_site_template'] = 'admin/base_site_default.html'

        ret['password_login_disabled'] = getattr(settings, 'HELUSERS_PASSWORD_LOGIN_DISABLED', False)

        return ret

    @never_cache
    def logout(self, request, extra_context=None):
        if request.session and request.session.get('social_auth_end_session_url'):
            logout_url = reverse('helusers:auth_logout')
            return HttpResponseRedirect(logout_url)
        return super().logout(request, extra_context)


# Django versions starting from 2.1 support overriding the default admin
# site, so monkeypatching is not needed.
if django.VERSION[0:2] < (2, 1):
    site = AdminSite()
    site._registry.update(admin.site._registry)
    default_admin_site = admin.site
    # Monkeypatch the default admin site with the custom one
    admin.site = site
    admin.sites.site = site



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.contrib.admin.apps import AdminConfig


class HelusersConfig(AppConfig):
    name = 'helusers'
    verbose_name = _("Helsinki Users")


class HelusersAdminConfig(AdminConfig):
    default_site = 'helusers.admin_site.AdminSite'



from django.conf import settings
from django.utils.translation import ugettext as _
from rest_framework import exceptions
from rest_framework_jwt.authentication import JSONWebTokenAuthentication
from rest_framework_jwt.settings import api_settings

from .user_utils import get_or_create_user


def patch_jwt_settings():
    """Patch rest_framework_jwt authentication settings from allauth"""
    defaults = api_settings.defaults
    defaults['JWT_PAYLOAD_GET_USER_ID_HANDLER'] = (
        __name__ + '.get_user_id_from_payload_handler')

    if 'allauth.socialaccount' not in settings.INSTALLED_APPS:
        return

    from allauth.socialaccount.models import SocialApp
    try:
        app = SocialApp.objects.get(provider='helsinki')
    except SocialApp.DoesNotExist:
        return

    defaults['JWT_SECRET_KEY'] = app.secret
    defaults['JWT_AUDIENCE'] = app.client_id

# Disable automatic settings patching for now because it breaks Travis.
# patch_jwt_settings()


class JWTAuthentication(JSONWebTokenAuthentication):
    def authenticate_credentials(self, payload):
        user = get_or_create_user(payload)

        if user and not user.is_active:
            msg = _('User account is disabled.')
            raise exceptions.AuthenticationFailed(msg)

        return user


def get_user_id_from_payload_handler(payload):
    return payload.get('sub')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-12 08:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ADGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=200)),
                ('display_name', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='ADGroupMapping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ad_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='helusers.ADGroup')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ad_groups', to='auth.Group')),
            ],
            options={
                'verbose_name': 'AD group mapping', 'verbose_name_plural': 'AD group mappings'
            },
        ),
        migrations.AlterUniqueTogether(
            name='adgroupmapping',
            unique_together=set([('group', 'ad_group')]),
        ),
    ]



# -*- coding: utf-8 -*-
import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-helusers',
    version='0.5.6',
    packages=['helusers'],
    include_package_data=True,
    license='BSD License',
    description='Django app for the user infrastructure of the City of Helsinki',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/City-of-Helsinki/django-helusers',
    author='City of Helsinki',
    author_email='dev@hel.fi',
    install_requires=[
        'Django',
        'drf-oidc-auth>=0.9',
        'requests',
        'python-jose',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from allauth.socialaccount import providers
from allauth.socialaccount.providers.base import ProviderAccount
from allauth.socialaccount.providers.oauth2.provider import OAuth2Provider

from helusers.utils import uuid_to_username


class HelsinkiAccount(ProviderAccount):
    def get_profile_url(self):
        return self.account.extra_data.get('html_url')

    def get_avatar_url(self):
        return self.account.extra_data.get('avatar_url')

    def to_str(self):
        dflt = super(HelsinkiAccount, self).to_str()
        return self.account.extra_data.get('name', dflt)


class HelsinkiProvider(OAuth2Provider):
    id = 'helsinki'
    name = 'City of Helsinki employees'
    package = 'helusers.providers.helsinki'
    account_class = HelsinkiAccount

    def extract_uid(self, data):
        return str(data['uuid'])

    def extract_common_fields(self, data):
        ret = data.copy()
        ret['username'] = uuid_to_username(data['uuid'])
        return ret

    def get_default_scope(self):
        return ['read']


providers.registry.register(HelsinkiProvider)



import urllib.parse as urlparse

from social_core.backends.open_id_connect import OpenIdConnectAuth
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch


class TunnistamoOIDCAuth(OpenIdConnectAuth):
    name = 'tunnistamo'
    OIDC_ENDPOINT = 'https://api.hel.fi/sso/openid'
    END_SESSION_URL = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Allow OIDC endpoint to be overridden through local settings
        self.OIDC_ENDPOINT = self.setting('OIDC_ENDPOINT', self.OIDC_ENDPOINT)

    def get_end_session_url(self, request, id_token):
        url = self.END_SESSION_URL or \
            self.oidc_config().get('end_session_endpoint')

        params = dict(id_token_hint=id_token)
        try:
            post_logout_url = reverse('helusers:auth_logout_complete')
        except NoReverseMatch:
            post_logout_url = None
        if post_logout_url:
            params['post_logout_redirect_uri'] = request.build_absolute_uri(post_logout_url)

        try:
            # Add the params to the end_session URL, which might have query params already
            url_parts = list(urlparse.urlparse(url))
            query = dict(urlparse.parse_qsl(url_parts[4]))
            query.update(params)
            url_parts[4] = urlparse.urlencode(query)
            end_session_url = urlparse.urlunparse(url_parts)
        except Exception:
            end_session_url = None

        return end_session_url



import uuid
import logging
from django.db import models, transaction
from django.contrib.auth.models import Group, AbstractUser as DjangoAbstractUser
from django.utils.translation import ugettext_lazy as _

from .utils import uuid_to_username


logger = logging.getLogger(__name__)


class ADGroup(models.Model):
    # Because AD group names are case insensitive, name is saved as lowercase.
    name = models.CharField(max_length=200, db_index=True)
    display_name = models.CharField(max_length=200)

    def save(self, *args, **kwargs):
        self.name = self.name.lower()
        return super().save(*args, **kwargs)

    def __str__(self):
        return self.display_name


class ADGroupMapping(models.Model):
    group = models.ForeignKey(Group, db_index=True, on_delete=models.CASCADE,
                              related_name='ad_groups')
    ad_group = models.ForeignKey(ADGroup, db_index=True, on_delete=models.CASCADE,
                                 related_name='groups')

    def __str__(self):
        return '%s -> %s' % (self.ad_group, self.group)

    class Meta:
        unique_together = (('group', 'ad_group'),)
        verbose_name = _("AD group mapping")
        verbose_name_plural = _("AD group mappings")


class AbstractUser(DjangoAbstractUser):
    uuid = models.UUIDField(unique=True)
    department_name = models.CharField(max_length=50, null=True, blank=True)
    ad_groups = models.ManyToManyField(ADGroup, blank=True)

    def save(self, *args, **kwargs):
        self.clean()
        return super(AbstractUser, self).save(*args, **kwargs)

    def clean(self):
        self._make_sure_uuid_is_set()
        if not self.username:
            self.set_username_from_uuid()

    def _make_sure_uuid_is_set(self):
        if self.uuid is None:
            self.uuid = uuid.uuid1()

    def set_username_from_uuid(self):
        self._make_sure_uuid_is_set()
        self.username = uuid_to_username(self.uuid)

    def get_display_name(self):
        if self.first_name and self.last_name:
            return '{0} {1}'.format(self.first_name, self.last_name).strip()
        else:
            return self.email

    def get_short_name(self):
        if self.first_name:
            return self.first_name
        return self.email

    def get_username(self):
        if not self.username or self.username.startswith('u-'):
            return self.email
        return self.username

    def natural_key(self):
        return (str(self.uuid),)

    def sync_groups_from_ad(self):
        """Determine which Django groups to add or remove based on AD groups."""

        ad_list = ADGroupMapping.objects.values_list('ad_group', 'group')
        mappings = {ad_group: group for ad_group, group in ad_list}

        user_ad_groups = set(self.ad_groups.filter(groups__isnull=False).values_list(flat=True))
        all_mapped_groups = set(mappings.values())
        old_groups = set(self.groups.filter(id__in=all_mapped_groups).values_list(flat=True))
        new_groups = set([mappings[x] for x in user_ad_groups])

        groups_to_delete = old_groups - new_groups
        if groups_to_delete:
            self.groups.remove(*groups_to_delete)
        groups_to_add = new_groups - old_groups
        if groups_to_add:
            self.groups.add(*groups_to_add)

    @transaction.atomic
    def update_ad_groups(self, ad_group_names):
        # Lock the User object to prevent races
        user = type(self).objects.select_for_update().get(id=self.id)

        # Make sure there's an ADGroup object for each group
        lower_names = [x.lower() for x in ad_group_names]
        ad_groups = {x.name.lower(): x for x in ADGroup.objects.filter(name__in=lower_names)}
        for name in ad_group_names:
            n = name.lower()
            if n not in ad_groups:
                ad_groups[n] = ADGroup.objects.create(name=n, display_name=name)

        # Update user's groups
        new_ad_groups = set([x.id for x in ad_groups.values()])
        old_ad_groups = set([x.id for x in user.ad_groups.all()])
        groups_to_add = new_ad_groups - old_ad_groups
        if groups_to_add:
            user.ad_groups.add(*groups_to_add)
        groups_to_remove = old_ad_groups - new_ad_groups
        if groups_to_remove:
            user.ad_groups.remove(*groups_to_remove)

        user.sync_groups_from_ad()

    def __str__(self):
        if self.first_name and self.last_name:
            return '%s %s (%s)' % (self.last_name, self.first_name, self.email)
        else:
            return self.email

    class Meta:
        abstract = True
        ordering = ('id',)
